"""
broadway.app
~~~~~~~~~~~

A factory to generate flask applications using a standard extension pattern.

:copyright: (c) 2017 Daniel Knell
:license: MIT, see LICENSE for more details.
"""

import copy

import flask
import twelvefactor
import werkzeug.utils

import broadway.config


class Factory(object):
    """
    The factory class creates flask applications. It is an implementation of
    the `application factory`_ pattern.


    .. _application factory: http://flask.pocoo.org/docs/patterns/appfactories/

    A factory performs the following tasks:

    * Create a new flask application and configure the it using the environment
    * Setup extensions to use the application
    * Load any blueprints and register them on the application

    The factory object is callable, any arguments are passed to the
    :class:`~flask.Flask` constructor.

    .. code-block:: python

        factory = Factory()

        factory.add_config_schema({
            'DEBUG': {
                'type': bool,
                'default': False,
            },
            'SECRET_KEY': str,
        })

        factory.add_extension('my_broadway_extension')

        factory.add_blueprint('myapp.views:blueprint')

        app = factory(__name__)

        if __name__ == '__main__':
            app.run()
    """

    def __init__(self):
        self._extensions = []
        self._blueprints = []
        self._schema = {}

    def add_config_schema(self, schema):
        """
        Add keys to the config schema

        .. code-block:: python

            factory = Factory()

            factory.add_config_schema({
                'DEBUG': {
                    'type': bool,
                    'default': False,
                },
                'SECRET_KEY': str,
            })

        :param schema: a dictionary to merge into the config schema
        :type schema: dict
        """
        self._schema.update(copy.deepcopy(schema))

    def add_blueprint(self, blueprint, **options):
        """
        Specify a blueprint to be registered with the application.

        Additional options will be passed to :meth:`~Flask.register_blueprint`
        when the application is created.

        .. code-block:: python

            factory = Factory()

            factory.add_blueprint('myapp.views:blueprint', url_prefix='/foo')

        :param blueprint: import path to blueprint object
        :type blueprint: str
        :param options: options to pass to the blueprint
        :type options: dict
        """
        instance = werkzeug.utils.import_string(blueprint)
        self._blueprints.append((instance, options))

    def add_blueprints(self, *blueprints):
        """
        Specify multiple blueprints to be registered with the application.

        If you need to pass extra arguments to :meth:`~Flask.register_blueprint`
        then see :meth:`add_blueprint` instead.

        .. code-block:: python

            factory = Factory()

            factory.add_blueprints(
                'myapp.views.auth:blueprint',
                'myapp.views.general:blueprint'
            )

        :param blueprints: a list of import paths
        :type blueprints: list
        """
        for blueprint in blueprints:
            self.add_blueprint(blueprint)

    def add_extension(self, extension):
        """
        Specify a broadway extension to initialise

        .. code-block:: python

            factory = Factory()

            factory.add_extension('broadway_sqlalchemy')

        :param extension: import path to extension
        :type extension: str
        """
        instance = werkzeug.utils.import_string(extension)

        if hasattr(instance, 'register'):
            instance.register(self)

        self._extensions.append(instance)

    def add_extensions(self, *extensions):
        """
        Specify multiple broadway extension to initialise

        .. code-block:: python

            factory = Factory()

            factory.add_extensions(
                'broadway_sqlalchemy',
                'some_extension.intergrations.broadway'
            )

        :param extensions: a list of import paths
        :type extensions: list
        """
        for extension in extensions:
            self.add_extension(extension)

    def __call__(self, *args, **kwargs):
        """
        Create a flask object, all arguments are passed to the :type:`Flask`
        constructor, extensions initialised, and blueprints registered.
        """
        instance = flask.Flask(*args, **kwargs)
        instance.config.update(twelvefactor.config(self._schema))

        for extension in self._extensions:
            if hasattr(extension, 'init_app'):
                extension.init_app(instance)

        for blueprint, options in self._blueprints:
            instance.register_blueprint(blueprint, **options)

        return instance


def factory():
    """
    Create a :class:`Factory` preconfigured with the flask configuration
    schema, whitenoise asset serving, and simple error handling.

    .. code-block:: python

        factory = broadway.factory()

        application = factory('yourapplication')


    """
    instance = Factory()
    instance.add_config_schema(broadway.config.SCHEMA)

    return instance
