"""
broadway.errors
~~~~~~~~~~~~~

An extension to provide a standard error handler, the handler will attempt to
return a response containing the content of a rendered template based on the
exceptions error code.

:copyright: (c) 2017 Daniel Knell
:license: MIT, see LICENSE for more details.
"""

import flask
import jinja2.exceptions
import werkzeug.exceptions

TEMPLATE = 'errors/{code}.html'


def handle_http_exception(error):
    """
    Attempts to render a template based on the `code` of a `HTTPException`. The
    format string `errors/{code}.html` is used to generate the filename of
    the template.

    e.g. A `NotFound` exception will attempt to render `errors/404.html`.

    If no template is found the standard flask error response is returned.
    """
    template = TEMPLATE.format(code=error.code)

    try:
        return flask.render_template(template, exception=error), error.code
    except jinja2.exceptions.TemplateNotFound:
        pass

    return error


def init_app(application):
    """
    Associates the error handler
    """
    for code in werkzeug.exceptions.default_exceptions:
        application.register_error_handler(code, handle_http_exception)
